@@ -13,14 +13,15 @@ from itertools import islice, chain, tee, compress, repeat
 from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping
 from coba.backports import Literal
 
-from coba            import pipes, primitives, CobaContext
+from coba            import pipes, primitives
 from coba.random     import CobaRandom
 from coba.exceptions import CobaException
 from coba.statistics import iqr
 from coba.utilities  import peek_first
-from coba.primitives import ScaleReward, BinaryReward, SequenceReward, BatchReward
+from coba.primitives import ScaleReward, BinaryReward, SequenceReward, BatchReward, IPSReward
 from coba.primitives import Feedback, BatchFeedback
 from coba.learners   import Learner, SafeLearner
+from coba.pipes      import Filter
 
 from coba.environments.primitives import EnvironmentFilter, Interaction
 
@@ -38,7 +39,26 @@ class Slice(pipes.Slice, EnvironmentFilter):
 
 class Shuffle(pipes.Shuffle, EnvironmentFilter):
     """Shuffle a sequence of Interactions in an Environment."""
-    pass
+    
+    def filter(self, interactions: Iterable[Interaction]) -> Sequence[Any]:
+        first, interactions = peek_first(interactions)
+
+        if 'action' in first and 'reward' in first:
+            #this is here because otherwise offpolicy evaluation can give a
+            #very biased estimate when seeds are the same. To see this run.
+                # import numpy as np
+                # from coba import CobaRandom
+
+                # N    = 1_000_000
+                # seed = 9
+
+                # R1 = CobaRandom(seed).randoms(N)
+                # R2 = CobaRandom(seed).shuffle(R1)
+
+                # np.corrcoef(R1,R2)
+            yield from CobaRandom("AB").shuffle(list(super().filter(interactions)))
+        else:
+            yield from super().filter(interactions)
 
 class Reservoir(pipes.Reservoir, EnvironmentFilter):
     """Take a fixed number of random Interactions from an Environment."""
@@ -922,7 +942,7 @@ class Unbatch(EnvironmentFilter):
 
         first, interactions = peek_first(interactions)
         if not interactions: return []
-        
+
         batched_keys = [k for k,v in first.items() if isinstance(v,primitives.Batch) ]
 
         if not batched_keys:
@@ -1017,12 +1037,14 @@ class Chunk(EnvironmentFilter):
 
 class Logged(EnvironmentFilter):
 
-    def __init__(self, learner: Learner) -> None:
+    def __init__(self, learner: Learner, rewards:Literal["DM","IPS"] = "DM", seed:float = 1.23) -> None:
         self._learner = learner
+        self._rewards = rewards
+        self._seed    = seed
 
     @property
     def params(self) -> Mapping[str, Any]:
-        return {"learner": SafeLearner(self._learner).params, "logged":True}
+        return {"learner": SafeLearner(self._learner).params, "logged":True, "rewards": self._rewards, "log_seed":self._seed}
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
@@ -1033,64 +1055,29 @@ class Logged(EnvironmentFilter):
         if 'context' not in first or 'actions' not in first or 'rewards' not in first:
             raise CobaException("We were unable to create a logged representation of the interaction.")
 
-        learner = SafeLearner(copy.deepcopy(self._learner))
+        #Avoid circular dependency
+        from coba.experiments.tasks import SimpleEvaluation
 
-        #In theory all of the code below could be replaced by:
+        I1,I2 = tee(interactions,2)
+        flat_int = Unbatch().filter(I1)
+        eval_log = SimpleEvaluation(record=['action','reward','probability']).process(copy.deepcopy(self._learner),I2,self._seed)
+        for interaction, log in zip(flat_int,eval_log):
+            out = interaction.copy()
+            out.update(log)
 
-            # I1,I2 = tee(interactions,2)
-            # for interaction, log in zip(I1, SimpleEvaluation(record=['action','reward','probability']).filter(I2)):
-            #     out = interaction.copy()
+            if self._rewards == "IPS":
+                actions = interaction.get('actions',[])
+                action  = actions.index(log['action']) if len(actions) > 0 else log['action']
+                out['rewards'] = IPSReward(log['reward'],action,log['probability'])
 
-            #     out['action']      = log['action']
-            #     out['reward']      = log['reward']
-            #     out['probability'] = log['probability']
-
-            #     yield out
-
-        #The reason it hasn't is because referencing SimpleEvalution creates a recursive dependency 
-
-        predict = learner.predict
-        learn   = learner.learn
-
-        learning_info = CobaContext.learning_info
-
-        def indexify(action,actions,discrete,batched):
-            if not discrete: return action
-            return [ A.index(a) for a,A in zip(action,actions) ] if batched else actions.index(action)
-
-        for interaction in interactions:
-
-            context = interaction['context']
-            actions = interaction['actions']
-            rewards = interaction['rewards']
-
-            batched = isinstance(context, primitives.Batch)
-            discrete = len(interaction.get('actions',[])) > 0
-
-            action,prob,info = predict(context, actions)
-
-            reward = rewards.eval(indexify(action,actions,discrete,batched))
-
-            learn(context, actions, action, reward, prob, **info)
-
-            new = interaction.copy()
-
-            new['action']      = action
-            new['probability'] = prob
-            new['reward']      = reward
-
-            if learning_info:
-                new.update(learning_info)
-                learning_info.clear()
-
-            yield new
+            yield out
 
 class Mutable(EnvironmentFilter):
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
-        
+
         first, interactions = peek_first(interactions)
-        
+
         if not interactions: return []
 
         first_is_dense   = isinstance(first['context'], primitives.Dense)
@@ -1103,13 +1090,13 @@ class Mutable(EnvironmentFilter):
                 new = interaction.copy()
                 new['context'] = new['context'].copy()
                 yield new
-                            
+
         elif first_is_dense:
             for interaction in interactions:
                 new = interaction.copy()
                 new['context'] = list(new['context'])
                 yield new
-        
+
         elif first_is_sparse:
             for interaction in interactions:
                 new = interaction.copy()
@@ -1119,3 +1106,7 @@ class Mutable(EnvironmentFilter):
         elif first_is_value:
             for interaction in interactions:
                 yield interaction.copy()
+
+class MappingToInteraction(Filter[Iterable[Mapping], Iterable[Interaction]]):
+    def filter(self, items: Iterable[Mapping]) -> Iterable[Interaction]:
+        yield from map(Interaction.from_dict,items)
