@@ -2,7 +2,7 @@ from abc import abstractmethod, ABC
 from typing import Any, Union, Iterable, Sequence, Mapping, overload
 
 from coba.primitives import Context, Action, Actions
-from coba.primitives import Reward, SequenceReward, Feedback, SequenceFeedback
+from coba.primitives import Reward, SequenceReward, Feedback, SequenceFeedback, IPSReward
 from coba.pipes import Source, SourceFilters, Filter
 from coba.exceptions import CobaException
 
@@ -15,11 +15,22 @@ class Interaction(dict):
     def extra(self) -> Mapping[str,Any]:
         return { k:self[k] for k in self.keys()-self.keywords}
 
+    @staticmethod
+    def from_dict(kwargs_dict: Mapping[str, Any]) -> 'Interaction':
+        if 'feedbacks' in kwargs_dict:
+            sub_class =  GroundedInteraction
+        elif 'rewards' in kwargs_dict:
+            sub_class = SimulatedInteraction
+        else:
+            sub_class =  LoggedInteraction
+
+        return sub_class(**kwargs_dict)
+
 class SimulatedInteraction(Interaction):
-    """Simulated data that provides labels for every possible action."""
+    """Simulated data that provides rewards for every possible action."""
     __slots__=()
     keywords = {'type', 'context', 'actions', 'rewards'}
-    
+
     def __init__(self,
         context : Context,
         actions : Actions,
@@ -46,7 +57,7 @@ class SimulatedInteraction(Interaction):
         if kwargs: self.update(kwargs)
 
 class GroundedInteraction(Interaction):
-    """Logged data providing a label for one action."""
+    """A grounded interaction based on Interaction Grounded Learning which feedbacks instead of rewards."""
     __slots__=()
     keywords = {'type', 'context', 'actions', 'rewards', 'feedbacks'}
 
@@ -74,7 +85,7 @@ class GroundedInteraction(Interaction):
         if kwargs: self.update(kwargs)
 
 class LoggedInteraction(Interaction):
-    """Logged data that describes an interaction where the choice was already made."""
+    """A logged interaction with an action, reward and optional probability."""
     __slots__ = ()
     keywords = {'type', 'context', 'action', 'reward', 'probability', 'actions', 'rewards'}
 
@@ -109,19 +120,17 @@ class LoggedInteraction(Interaction):
             **kwargs : Any additional information.
         """
 
-        if 'actions' in kwargs and 'probability' in kwargs and 'rewards' not in kwargs:
-            probability       = kwargs['probability']
-            actions           = kwargs['actions']
-            kwargs['rewards'] = [int(a==action)*reward/probability for a in actions]
+        if isinstance(kwargs.get('rewards'),(list,tuple)):
+            self['rewards'] = SequenceReward(kwargs.pop('rewards'))
+        elif kwargs.get('rewards') is None and kwargs.get('actions') is not None:
+            ips_action = kwargs['actions'].index(action) if isinstance(kwargs['actions'],(list,tuple)) else action
+            self['rewards'] = IPSReward(reward,ips_action,kwargs.get('probability'))
 
         self['context'] = context
         self['action']  = action
         self['reward']  = reward
 
-        if 'rewards' in kwargs and isinstance(kwargs['rewards'],(list,tuple)):
-            kwargs['rewards'] = SequenceReward(kwargs['rewards'])
-
-        if kwargs: self.update(kwargs)
+        if kwargs: self.update({k:v for k,v in kwargs.items() if v is not None})
 
 class EnvironmentFilter(Filter[Iterable[Interaction],Iterable[Interaction]], ABC):
     """A filter that can be applied to an Environment."""
