@@ -19,7 +19,6 @@ from coba.primitives import Batch
 from coba.statistics import percentile
 from coba.utilities import PackageChecker, peek_first
 
-
 class LearnerTask(ABC):
     """A task which describes a Learner."""
 
@@ -55,22 +54,36 @@ class EvaluationTask(ABC):
     """A task which evaluates a Learner on an Environment."""
 
     @abstractmethod
-    def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Mapping[Any,Any]]:
+    def process(self, learner: Learner, interactions: Iterable[Interaction], seed:float = None) -> Iterable[Mapping[Any,Any]]:
         """Process the EvaluationTask.
 
         Args:
             learner: The Learner that we wish to evaluate.
             interactions: The Interactions which we wish to evaluate on.
+            seed: An optional argument to seed randomness during evaluation.
 
         Returns:
-            An iterable of evaluation statistics (for online evaluation there should be one dict per interaction).
+            An iterable of evaluation results (e.g., for online evaluation there should be one dict per interaction).
         """
         ...
 
+class SeededEvaluation(EvaluationTask):
+
+    def __init__(self, evaluator: EvaluationTask, seed:float = None):
+        self._evaluator = evaluator
+        self._seed      = seed
+
+    def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Mapping[Any, Any]]:
+        try:
+            return self._evaluator.process(learner, interactions, self._seed)
+        except TypeError as ex:
+            if 'process()' not in str(ex) or '3' not in str(ex) or '4' not in str(ex): raise
+            return self._evaluator.process(learner, interactions)
+
 class SimpleEvaluation(EvaluationTask):
 
     def __init__(self, 
-        record: Sequence[Literal['reward','rank','regret','time','probability','action', 'context', 'ope_loss']] = ['reward'],
+        record: Sequence[Literal['reward','rank','regret','time','probability','action','actions', 'context', 'ope_loss', 'rewards']] = ['reward'],
         learn: bool = True,
         predict: bool = True) -> None:
         """
@@ -90,15 +103,9 @@ class SimpleEvaluation(EvaluationTask):
             # https://vowpalwabbit.org/docs/vowpal_wabbit/python/latest/tutorials/off_policy_evaluation.html
             PackageChecker.vowpalwabbit('SimpleEvaluation.__init__')
 
-    def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Mapping[Any,Any]]:
-
-        learner = SafeLearner(learner)
-
-        record_prob    = 'probability' in self._record
-        record_time    = 'time'        in self._record
-        record_action  = 'action'      in self._record
-        record_context = 'context'     in self._record
-        record_ope_loss = 'ope_loss'     in self._record
+    def process(self, learner: Learner, interactions: Iterable[Interaction], seed: float = 1) -> Iterable[Mapping[Any,Any]]:
+        
+        learner = SafeLearner(learner, seed)
 
         first, interactions = peek_first(interactions)
 
@@ -107,10 +114,20 @@ class SimpleEvaluation(EvaluationTask):
 
         discrete = first and len(first.get('actions',[])) > 0
 
+        if not discrete:
+            DISCRETE_ONLY_METRICS = {'rank', 'rewards'}
+            for metric in set(self._record).intersection(DISCRETE_ONLY_METRICS):
+                warnings.warn(f"The {metric} metric can only be calculated for discrete environments")
+
         learning_info = CobaContext.learning_info
 
-        if 'rank' in self._record and not discrete:
-            warnings.warn(f"The rank metric can only be calculated for discrete environments")
+        record_prob     = 'probability' in self._record
+        record_time     = 'time'        in self._record
+        record_action   = 'action'      in self._record
+        record_context  = 'context'     in self._record
+        record_ope_loss = 'ope_loss'    in self._record
+        record_rewards  = 'rewards'     in self._record and discrete
+        record_actions  = 'actions'     in self._record
 
         calc_rank   = 'rank'   in self._record and discrete
         calc_reward = 'reward' in self._record
@@ -160,19 +177,25 @@ class SimpleEvaluation(EvaluationTask):
                 reward   = rewards.eval(_action)
                 feedback = feedbacks.eval(_action) if feedbacks else reward
 
-                if record_time  : out['predict_time'] = predict_time
-                if record_prob  : out['probability']  = prob
-                if record_action: out['action']       = action
-                if feedbacks    : out['feedback']     = feedback
+                if record_time   : out['predict_time'] = predict_time
+                if record_prob   : out['probability']  = prob
+                if record_action : out['action']       = action
+                if record_actions: out['actions']      = actions
+                if feedbacks     : out['feedback']     = feedback
+
+                if not batched:
+                    if record_rewards: out['rewards'] = list(map(rewards.eval, range(len(actions))))
+                else:
+                    if record_rewards: out['rewards'] = list(zip(*[rewards.eval([a]*len(context)) for a in range(len(actions[0]))]))
 
                 if not batched:
                     if calc_reward : out['reward'] = get_reward(reward)
                     if calc_regret : out['regret'] = get_regret(reward, rewards)
                     if calc_rank   : out['rank'  ] = get_rank  (reward, rewards, len(actions))
                 elif batched:
-                    if calc_reward : out['reward'] = mean(map(get_reward,reward))
-                    if calc_regret : out['regret'] = mean(map(get_regret,reward,rewards))
-                    if calc_rank   : out['rank'  ] = mean(map(get_rank  ,reward,rewards,len(actions)))
+                    if calc_reward : out['reward'] = list(map(get_reward,reward))
+                    if calc_regret : out['regret'] = list(map(get_regret,reward,rewards))
+                    if calc_rank   : out['rank'  ] = list(map(get_rank  ,reward,rewards,len(actions)))
 
             if is_learn:
                 if is_off_policy_learn:
@@ -207,7 +230,19 @@ class SimpleEvaluation(EvaluationTask):
                 out.update(learning_info)
                 learning_info.clear()
 
-            if out: yield out
+            if out:
+
+                # for off-policy evaluation purposes we need to log each reward even when batching.
+                # Therefore, we have two options:
+                #   (1) store a batched version of reward and keep out batched. With this option we
+                #       also need to modify Results so that it can plut batched versions of "reward".
+                #   (2) we flatten out and store a normal version of reward. We lose some information
+                #       downstream this way but we gain "reward" now playing nicely with all plots.
+                # Here we went with option (2). If changing to (1) be sure to update Result.
+                if batched:
+                    yield from ({k:v[i] for k,v in out.items()} for i in range(len(context)))
+                else:
+                    yield out
 
 class SimpleLearnerInfo(LearnerTask):
     """Describe a Learner using its name and hyperparameters."""
