@@ -1,15 +1,13 @@
 import re
-
-from sys import platform
 from itertools import repeat, compress
+from sys import platform
 from typing import Any, Dict, Union, Sequence, Mapping, Optional, Tuple
+
 from coba.backports import Literal
-
 from coba.exceptions import CobaException
-from coba.utilities import PackageChecker
+from coba.learners.primitives import Learner, PMF
 from coba.primitives import Sparse, Context, Action, Actions
-
-from coba.learners.primitives import Learner, Probs
+from coba.utilities import PackageChecker
 
 Feature       = Union[str,int,float]
 Features      = Union[Feature, Sequence[Feature], Dict[str,Union[int,float]]]
@@ -17,6 +15,8 @@ Namespaces    = Dict[str,Features]
 VW_Features   = Sequence[Union[str,int,Tuple[Union[str,int],Union[int,float]],Dict[str,Union[int,float]]]]
 VW_Namespaces = Dict[str,VW_Features]
 
+DEFAULT_NAMESPACE_INTERACTIONS = (1, 'a', 'ax', 'axx')
+
 class VowpalMediator:
     """A class to handle all communication between Coba and VW."""
 
@@ -94,6 +94,15 @@ class VowpalMediator:
 
         return ex
 
+    def finish(self):
+        if self.is_initialized:
+            self._vw.finish()
+    
+    # override to transform example before they get pushed down to VW
+    # i.e. change namespaces from the default 'x' in shared to any other namespace
+    def transform_example(self, vw_shared, vw_uniques, labels):
+        pass
+
     def make_examples(self, shared: Namespaces, uniques: Sequence[Namespaces], labels:Optional[Sequence[str]]) -> Sequence[Any]:
         """Create a list of VW examples.
 
@@ -107,6 +116,8 @@ class VowpalMediator:
         vw_shared  = dict(self._prep_namespaces(shared))
         vw_uniques = list(map(dict,map(self._prep_namespaces,uniques)))
 
+        self.transform_example(vw_shared, vw_uniques, labels)
+
         examples = []
         for vw_unique, label in zip(vw_uniques,labels):
             ex = self._example_init(self._vw, None, self._label_type)
@@ -130,7 +141,7 @@ class VowpalMediator:
                 elif feats.__class__ is int or feats.__class__ is float:
                     yield (ns, [(self._get_namespace_keys(ns,1)[0], feats)])
                 else:
-                    feats = feats.items() if isinstance(feats, Sparse) else enumerate(feats,self._get_namespace_keys(ns,len(feats)))
+                    feats = feats.items() if isinstance(feats, Sparse) else zip(self._get_namespace_keys(ns,len(feats)),feats)
                     yield (ns, [f"{k}={v}" if v.__class__ is str else (k, v) for k,v in feats if v!= 0])
         else: #pragma: no cover
             #the strange type checks below were faster than traditional methods when performance testing
@@ -142,7 +153,19 @@ class VowpalMediator:
                 elif feats.__class__ is int or feats.__class__ is float:
                     yield (ns, [(self._get_namespace_keys(ns,1)[0], feats)])
                 elif isinstance(feats,Sparse):
-                    yield (ns, feats.copy())
+                    no_str = not any(map(isinstance,feats.values(),repeat(str))) # most-performant (still 2x slower than no check)
+                    #no_str = str not in set(map(attrgetter("__class__"),feats.values()))
+                    #no_str = not any(map(is_,map(attrgetter("__class__"),feats.values()),repeat(str)))
+                    #no_str = True
+                    if no_str:
+                        yield (ns, feats.copy())
+                    else:
+                        new = feats.copy()
+                        for k,v in feats.items():
+                            if isinstance(v,str):
+                                new[f"{k}={v}"] = 1
+                                del new[k]
+                        yield (ns,new)
                 else:
                     d={}
 
@@ -185,45 +208,39 @@ class VowpalLearner(Learner):
 
     @staticmethod
     def make_args(
-        options: Sequence[str],
-        noconstant: bool,
-        interactions: Sequence[str],
-        ignore_linear:Sequence[str],
-        seed: Optional[int],
-        **kwargs) -> str:
-        """Turn settings into a VW command line string.
-
-        Args:
-            options: A sequence of string values that represent VW CLI options.
-            noconstant: Indicates if constant term should be included in the VW features.
-            interactions: A sequence of namespace interactions to use during learning.
-            ignore_linear: A sequence of linear namespaces to ignore during learning.
-            seed: A random number generator seed to make sure VW behaves consistently.
-            kwargs: Any number of additional options to add to the arg string.
+            vw_kwargs: Dict[str, Any],
+            namespace_interactions: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+    ) -> Sequence[str]:
         """
+        Turn settings into VW command line arguments.
 
-        options = list(filter(None,options))
+         Args:
+            vw_kwargs: Keyword argument dict that gets translated into VW CLI arguments
+            namespace_interactions: A sequence of namespace interactions to use during learning.
+        """
+        args = []
+        vw_kwargs['quiet'] = vw_kwargs.get('quiet', True)
+
+        for k,v in vw_kwargs.items():
+            if v is not False and v is not None:
+                if not k.startswith("-"):
+                    k = ("-" if len(k)==1 else "--") + k
+                args.append(k if (v is True) else f"{k} {v}")
+
+        noconstant = sum([n for n in namespace_interactions if isinstance(n, (int, float))]) == 0
+        ignore_linear = set(['x', 'a']) - set(namespace_interactions)
+        interactions = [n for n in namespace_interactions if isinstance(n, str) and len(n) > 1]
 
         if noconstant:
-            options.append(f"--noconstant")
+            args.append(f"--noconstant")
 
         for interaction in interactions:
-            options.append(f"--interactions {interaction}")
+            args.append(f"--interactions {interaction}")
 
         for ignore in ignore_linear:
-            options.append(f"--ignore_linear {ignore}")
+            args.append(f"--ignore_linear {ignore}")
 
-        if seed is not None:
-            options.append(f"--random_seed {seed}")
-
-        kwargs['quiet'] = kwargs.get('quiet',True)
-
-        for k,v in kwargs.items():
-            if v is not False:
-                k = ("-" if len(k)==1 else "--") + k
-                options.append(k if (v is None or v == True) else f"{k} {v}")
-
-        return " ".join(options)
+        return args
 
     def __init__(self, args: str = "--cb_explore_adf --epsilon 0.05 --interactions ax --interactions axx --ignore_linear x --random_seed 1 --quiet", vw: VowpalMediator = None) -> None:
         """Instantiate a VowpalLearner.
@@ -260,7 +277,7 @@ class VowpalLearner(Learner):
     def params(self) -> Mapping[str, Any]:
         return {"family": "vw", 'args': self._args.replace("--quiet","").strip()}
 
-    def predict(self, context: Context, actions: Sequence[Action]) -> Probs:
+    def predict(self, context: Context, actions: Sequence[Action]) -> PMF:
 
         if not self._vw.is_initialized and self._adf:
             self._vw.init_learner(self._args, 4)
@@ -306,9 +323,9 @@ class VowpalLearner(Learner):
             index = self._vw.predict(self._vw.make_example(context, None))
             probs = [ int(i==index) for i in range(1,len(actions)+1) ]
 
-        return Probs(probs)
+        return PMF(probs)
 
-    def learn(self, context: Context, actions:Actions, index: int, reward: float, probability: float) -> None:
+    def learn(self, context: Context, actions:Actions, action: Action, reward: float, probability: float) -> None:
 
         if not self._vw.is_initialized and self._adf:
             self._vw.init_learner(self._args, 4)
@@ -319,6 +336,7 @@ class VowpalLearner(Learner):
         if not self._vw.is_initialized and not self._adf and not self._n_actions:
             raise CobaException("When using `cb` without `adf` predict must be called before learn to initialize the vw learner")
 
+        index   = actions.index(action)
         labels  = self._labels(actions, index, reward, probability)
         label   = labels[index]
 
@@ -333,6 +351,26 @@ class VowpalLearner(Learner):
     def _labels(self,actions,index,reward:float,prob:float) -> Sequence[Optional[str]]:
         return [ f"{i+1}:{round(-reward,5)}:{round(prob,5)}" if i == index else None for i in range(len(actions))]
 
+    def _finish(self):
+        try:
+            self._vw.finish()
+        except AttributeError:
+            #we do these checks because when mocking up for unittests
+            #it is possible that _vw actually isn't ever initialized
+            pass
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self._finish()
+
+    def finish(self):
+        self._finish()
+
+    def __del__(self):
+        self._finish()
+
 class VowpalEpsilonLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
         information on the types of exploration algorithms availabe in VW see `here`__.
@@ -341,25 +379,27 @@ class VowpalEpsilonLearner(VowpalLearner):
     """
 
     def __init__(self,
-        epsilon: float = 0.05,
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-        **kwargs) -> None:
+                 epsilon: float = 0.05,
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalEpsilonLearner.
 
         Args:
             epsilon: The probability that we will explore instead of exploit.
-            features: A list of namespaces and interactions  to use when learning reward functions.
+            features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary random numbers.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
         """
-
-        options       = [ "--cb_explore_adf", f"--epsilon {epsilon}" ]
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
-
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
+        vw_kwargs = {
+            "cb_explore_adf": True,
+            "epsilon": epsilon,
+            "random_seed": seed
+        }
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
 
 class VowpalSoftmaxLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
@@ -369,28 +409,31 @@ class VowpalSoftmaxLearner(VowpalLearner):
     """
 
     def __init__(self,
-        softmax: float=10,
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-
-        **kwargs) -> None:
+                 softmax: float=10,
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalSoftmaxLearner.
 
         Args:
             softmax: An exploration parameter with 0 indicating predictions should be completely random
                 and infinity indicating that predictions should be greedy. For more information see `lambda`__.
-            features: A list of namespaces and interactions  to use when learning reward functions.
+            features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary randomness.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
 
         __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
         """
-
-        options       = [ "--cb_explore_adf", "--softmax", f"--lambda {softmax}" ]
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
+        vw_kwargs = {
+            "cb_explore_adf": True,
+            "softmax": True,
+            "lambda": softmax,
+            "random_seed": seed
+        }
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
 
 class VowpalBagLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
@@ -400,26 +443,29 @@ class VowpalBagLearner(VowpalLearner):
     """
 
     def __init__(self,
-        bag: int = 5,
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-        **kwargs) -> None:
+                 bag: int = 5,
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalBagLearner.
 
         Args:
             bag: This value determines the number of policies which will be learned and must be greater
                 than 0. Each policy is trained using bootstrap aggregation, making each policy unique. During
                 prediction a random policy will be selected according to a uniform distribution and followed.
-            features: A list of namespaces and interactions  to use when learning reward functions.
+            features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary random numbers.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
         """
-
-        options       = [ "--cb_explore_adf", f"--bag {bag}" ]
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
+        vw_kwargs = {
+            "cb_explore_adf": True,
+            "bag": bag,
+            "random_seed": seed
+        }
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
 
 class VowpalCoverLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
@@ -436,25 +482,69 @@ class VowpalCoverLearner(VowpalLearner):
     """
 
     def __init__(self,
-        cover: int = 5,
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-        **kwargs) -> None:
+                 cover: int = 5,
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalCoverLearner.
 
         Args:
             cover: The number of policies which will be learned (must be greater than 0).
-            features: A list of namespaces and interactions  to use when learning reward functions.
+            features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary random numbers.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
         """
 
-        options       = [ "--cb_explore_adf", f"--cover {cover}" ]
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
+        vw_kwargs = {
+            "cb_explore_adf": True,
+            "cover": cover,
+            "random_seed": seed
+        }
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
+
+
+class VowpalRndLearner(VowpalLearner):
+    def __init__(self,
+                 rnd: int = 3,
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 epsilon: Optional[float] = 0.025,
+                 rnd_alpha: Optional[float] = None,
+                 rnd_invlambda: Optional[float] = None,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
+        """
+        Inspired by Random Network Distillation, this explorer constructs an auxiliary prediction problem whose
+        expected target value is zero and uses the prediction magnitude to construct a confidence interval. In the
+        contextual bandit case this is equivalent to a randomized approximation to the LinUCB bound.
+
+        For more information see the `wiki`__.
+        __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
+
+        Args:
+            rnd: Number of predictors
+            features: A list of namespaces and interactions  to use when learning reward functions
+            epsilon: Uniform exploration term for stabilization
+            rnd_alpha: Increase for more exploration on a repeated example
+            rnd_invlambda: Increase for more exploration on examples with new features/actions
+            seed: The seed used by VW to generate any necessary random numbers
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
+        """
+        vw_kwargs = {
+            "cb_explore_adf": True,
+            "rnd": rnd,
+            "epsilon": epsilon,
+            "rnd_alpha": rnd_alpha,
+            "rnd_invlambda": rnd_invlambda,
+            "random_seed": seed
+        }
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
 
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
 
 class VowpalRegcbLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
@@ -469,26 +559,34 @@ class VowpalRegcbLearner(VowpalLearner):
     """
 
     def __init__(self,
-        mode: Literal["optimistic","elimination"] = "elimination",
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-        **kwargs) -> None:
+                 mode: Literal["optimistic","elimination"] = "elimination",
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalRegcbLearner.
 
         Args:
             mode: Indicates whether exploration should only predict the optimal upper bound action or
                 should use an elimination technique to remove actions that no longer seem plausible
                 and pick randomly from the remaining actions.
-            features: A list of namespaces and interactions  to use when learning reward functions.
+            features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary random numbers.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
         """
 
-        options       = [ "--cb_explore_adf", "--regcb" if mode=="elimination" else "--regcbopt" ]
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
+        vw_kwargs = {
+            "cb_explore_adf": True,
+        }
+        if mode == "elimination":
+            vw_kwargs["regcb"] = True
+        else:
+            vw_kwargs["regcbopt"] = True
+        vw_kwargs["random_seed"] = seed
+
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
 
 class VowpalSquarecbLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
@@ -503,11 +601,11 @@ class VowpalSquarecbLearner(VowpalLearner):
     """
 
     def __init__(self,
-        mode: Literal["standard","elimination"] = "standard",
-        gamma_scale: float = 10,
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-        **kwargs) -> None:
+                 mode: Literal["standard","elimination"] = "standard",
+                 gamma_scale: float = 10,
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalSquarecbLearner.
 
         Args:
@@ -518,20 +616,20 @@ class VowpalSquarecbLearner(VowpalLearner):
                 as gamma in the original paper.
             features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary random numbers.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
         """
-
-        options = [
-            "--cb_explore_adf",
-            "--squarecb",
-            f"--gamma_scale {gamma_scale}",
-            "" if mode != "elimination" else "--elim"
-        ]
-
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
+        vw_kwargs = {
+            "cb_explore_adf": True,
+            "squarecb": True,
+            "gamma_scale": gamma_scale,
+            "random_seed": seed
+        }
+        if mode == "elimination":
+            vw_kwargs["elim"] = True
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
 
 class VowpalOffPolicyLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
@@ -545,19 +643,22 @@ class VowpalOffPolicyLearner(VowpalLearner):
     """
 
     def __init__(self,
-        features: Sequence[str] = [1,'a','ax','axx'],
-        seed: Optional[int] = 1,
-        **kwargs) -> None:
+                 features: Sequence[str] = DEFAULT_NAMESPACE_INTERACTIONS,
+                 seed: Optional[int] = 1,
+                 **kwargs) -> None:
         """Instantiate a VowpalOffPolicyLearner.
 
         Args:
-            features: A list of namespaces and interactions  to use when learning reward functions.
+            features: A list of namespaces and interactions to use when learning reward functions.
             seed: The seed used by VW to generate any necessary random numbers.
+            kwargs: Additional key-word args are passed on as VW CLI arguments (unless removed in the function).
         """
 
-        options       = ["--cb_adf"]
-        noconstant    = sum([f for f in features if isinstance(f,(int,float))]) == 0
-        ignore_linear = set(['x','a'])-set(features)
-        interactions  = [f for f in features if isinstance(f,str) and len(f) > 1]
-        vw            = kwargs.pop('vw',None)
-        super().__init__(VowpalLearner.make_args(options, noconstant, interactions, ignore_linear, seed, **kwargs), vw)
+        vw_kwargs = {
+            "cb_adf": True,
+            "random_seed": seed
+        }
+        vw = kwargs.pop('vw', None)
+        vw_kwargs.update(kwargs)
+        vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
+        super().__init__(vw_args_string, vw)
